@@ -1,6 +1,6 @@
 ;;; exwm-floating.el --- Floating Module for EXWM  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2015-2018 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2019 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -44,20 +44,70 @@ context of the corresponding buffer."
 context of the corresponding buffer."
   :type 'hook)
 
+(defvar exwm-floating--border-pixel nil
+  "Border pixel drawn around floating X windows.")
+
 (defcustom exwm-floating-border-color "navy"
   "Border color of floating windows."
-  :type 'color)
+  :type 'color
+  :initialize #'custom-initialize-default
+  :set (lambda (symbol value)
+         (set-default symbol value)
+         ;; Change border color for all floating X windows.
+         (exwm-floating--init-border)
+         (dolist (pair exwm--id-buffer-alist)
+           (with-current-buffer (cdr pair)
+             (when exwm--floating-frame
+               (xcb:+request exwm--connection
+                   (make-instance 'xcb:ChangeWindowAttributes
+                                  :window
+                                  (frame-parameter exwm--floating-frame
+                                                   'exwm-container)
+                                  :value-mask xcb:CW:BorderPixel
+                                  :border-pixel
+                                  exwm-floating--border-pixel)))))
+         (when exwm--connection
+           (xcb:flush exwm--connection))))
 
 (defcustom exwm-floating-border-width 1
   "Border width of floating windows."
-  :type 'integer)
+  :type '(integer
+          :validate (lambda (widget)
+                      (when (< (widget-value widget) 0)
+                        (widget-put widget :error "Border width is at least 0")
+                        widget)))
+  :initialize #'custom-initialize-default
+  :set (lambda (symbol value)
+         (let ((delta (- value exwm-floating-border-width))
+               container)
+           (set-default symbol value)
+           ;; Change border width for all floating X windows.
+           (dolist (pair exwm--id-buffer-alist)
+             (with-current-buffer (cdr pair)
+               (when exwm--floating-frame
+                 (setq container (frame-parameter exwm--floating-frame
+                                                  'exwm-container))
+                 (with-slots (x y)
+                     (xcb:+request-unchecked+reply exwm--connection
+                         (make-instance 'xcb:GetGeometry
+                                        :drawable container))
+                   (xcb:+request exwm--connection
+                       (make-instance 'xcb:ConfigureWindow
+                                      :window container
+                                      :value-mask
+                                      (logior xcb:ConfigWindow:X
+                                              xcb:ConfigWindow:Y
+                                              xcb:ConfigWindow:BorderWidth)
+                                      :border-width value
+                                      :x (- x delta)
+                                      :y (- y delta)))))))
+           (when exwm--connection
+             (xcb:flush exwm--connection)))))
 
 (defvar exwm-floating--border-colormap nil
   "Colormap used by the border pixel.
 
 This is also used by X window containers.")
-(defvar exwm-floating--border-pixel nil
-  "Border pixel drawn around floating X windows.")
 
 ;; Cursors for moving/resizing a window
 (defvar exwm-floating--cursor-move nil)
@@ -74,12 +124,15 @@ This is also used by X window containers.")
   "Calculate move/resize parameters [buffer event-mask x y width height].")
 
 (defvar exwm-workspace--current)
+(defvar exwm-workspace--frame-y-offset)
+(defvar exwm-workspace--window-y-offset)
 (defvar exwm-workspace--workareas)
 (declare-function exwm-layout--hide "exwm-layout.el" (id))
 (declare-function exwm-layout--iconic-state-p "exwm-layout.el" (&optional id))
 (declare-function exwm-layout--refresh "exwm-layout.el" ())
 (declare-function exwm-layout--show "exwm-layout.el" (id &optional window))
 (declare-function exwm-workspace--position "exwm-workspace.el" (frame))
+(declare-function exwm-workspace--update-offsets "exwm-workspace.el" ())
 
 (defun exwm-floating--set-allowed-actions (id tilling)
   "Set _NET_WM_ALLOWED_ACTIONS."
@@ -116,8 +169,8 @@ This is also used by X window containers.")
                           (get-buffer "*scratch*")))
                   (make-frame
                    `((minibuffer . ,(minibuffer-window exwm--frame))
-                     (left . ,(* window-min-width -100))
-                     (top . ,(* window-min-height -100))
+                     (left . ,(* window-min-width -10000))
+                     (top . ,(* window-min-height -10000))
                      (width . ,window-min-width)
                      (height . ,window-min-height)
                      (unsplittable . t))))) ;and fix the size later
@@ -129,6 +182,9 @@ This is also used by X window containers.")
          (y (slot-value exwm--geometry 'y))
          (width (slot-value exwm--geometry 'width))
          (height (slot-value exwm--geometry 'height)))
+    ;; Force drawing menu-bar & tool-bar.
+    (redisplay t)
+    (exwm-workspace--update-offsets)
     (exwm--log "Floating geometry (original): %dx%d%+d%+d" width height x y)
     ;; Save frame parameters.
     (set-frame-parameter frame 'exwm-outer-id outer-id)
@@ -214,7 +270,9 @@ This is also used by X window containers.")
            (frame-width (+ width (- (frame-pixel-width frame)
                                     (- (elt edges 2) (elt edges 0)))))
            (frame-height (+ height (- (frame-pixel-height frame)
-                                      (- (elt edges 3) (elt edges 1)))))
+                                      (- (elt edges 3) (elt edges 1)))
+                            ;; Use `frame-outer-height' in the future.
+                            exwm-workspace--frame-y-offset))
            (floating-mode-line (plist-get exwm--configurations
                                           'floating-mode-line))
            (floating-header-line (plist-get exwm--configurations
@@ -239,7 +297,7 @@ This is also used by X window containers.")
                                     'floating-header-line))
                  exwm--mwm-hints-decorations)
             (setq header-line-format nil)
-          ;; The header-line need to be hidden in floating header.
+          ;; The header-line need to be hidden in floating mode.
           (setq frame-height (- frame-height (window-header-line-height
                                               (frame-root-window frame)))
                 header-line-format nil)))
@@ -250,8 +308,8 @@ This is also used by X window containers.")
                          :depth 0
                          :wid frame-container
                          :parent exwm--root
-                         :x (- x (elt edges 0))
-                         :y (- y (elt edges 1))
+                         :x x
+                         :y (- y exwm-workspace--window-y-offset)
                          :width width
                          :height height
                          :border-width
@@ -323,7 +381,7 @@ This is also used by X window containers.")
   (with-current-buffer (exwm--id->buffer id)
     (run-hooks 'exwm-floating-setup-hook))
   ;; Redraw the frame.
-  (redisplay))
+  (redisplay t))
 
 (defun exwm-floating--unset-floating (id)
   "Make window ID non-floating."
@@ -342,7 +400,7 @@ This is also used by X window containers.")
         (xcb:+request exwm--connection
             (make-instance 'xcb:ChangeWindowAttributes
                            :window id :value-mask xcb:CW:EventMask
-                           :event-mask exwm--client-event-mask))
+                           :event-mask (exwm--get-client-event-mask)))
         ;; Reparent the floating frame back to the root window.
         (let ((frame-id (frame-parameter exwm--floating-frame 'exwm-outer-id))
               (frame-container (frame-parameter exwm--floating-frame
@@ -466,7 +524,7 @@ This is also used by X window containers.")
                 height (frame-pixel-height frame))
           (unless type
             ;; Determine the resize type according to the pointer position
-            ;; Clicking the center 1/3 part to resize has not effect
+            ;; Clicking the center 1/3 part to resize has no effect
             (setq x (/ (* 3 win-x) (float width))
                   y (/ (* 3 win-y) (float height))
                   type (cond ((and (< x 1) (< y 1))
@@ -486,91 +544,91 @@ This is also used by X window containers.")
           (cond ((= type xcb:ewmh:_NET_WM_MOVERESIZE_MOVE)
                  (setq cursor exwm-floating--cursor-move
                        exwm-floating--moveresize-calculate
-                       `(lambda (x y)
-                          (vector ,buffer-or-id
-                                  ,(eval-when-compile
-                                     (logior xcb:ConfigWindow:X
-                                             xcb:ConfigWindow:Y))
-                                  (- x ,win-x) (- y ,win-y) 0 0))))
+                       (lambda (x y)
+                         (vector buffer-or-id
+                                 (eval-when-compile
+                                   (logior xcb:ConfigWindow:X
+                                           xcb:ConfigWindow:Y))
+                                 (- x win-x) (- y win-y) 0 0))))
                 ((= type xcb:ewmh:_NET_WM_MOVERESIZE_SIZE_TOPLEFT)
                  (setq cursor exwm-floating--cursor-top-left
                        exwm-floating--moveresize-calculate
-                       `(lambda (x y)
-                          (vector ,buffer-or-id
-                                  ,(eval-when-compile
-                                     (logior xcb:ConfigWindow:X
-                                             xcb:ConfigWindow:Y
-                                             xcb:ConfigWindow:Width
-                                             xcb:ConfigWindow:Height))
-                                  (- x ,win-x) (- y ,win-y)
-                                  (- ,(+ root-x width) x)
-                                  (- ,(+ root-y height) y)))))
+                       (lambda (x y)
+                         (vector buffer-or-id
+                                 (eval-when-compile
+                                   (logior xcb:ConfigWindow:X
+                                           xcb:ConfigWindow:Y
+                                           xcb:ConfigWindow:Width
+                                           xcb:ConfigWindow:Height))
+                                 (- x win-x) (- y win-y)
+                                 (- (+ root-x width) x)
+                                 (- (+ root-y height) y)))))
                 ((= type xcb:ewmh:_NET_WM_MOVERESIZE_SIZE_TOP)
                  (setq cursor exwm-floating--cursor-top
                        exwm-floating--moveresize-calculate
-                       `(lambda (x y)
-                          (vector ,buffer-or-id
-                                  ,(eval-when-compile
-                                     (logior xcb:ConfigWindow:Y
-                                             xcb:ConfigWindow:Height))
-                                  0 (- y ,win-y) 0 (- ,(+ root-y height) y)))))
+                       (lambda (_x y)
+                         (vector buffer-or-id
+                                 (eval-when-compile
+                                   (logior xcb:ConfigWindow:Y
+                                           xcb:ConfigWindow:Height))
+                                 0 (- y win-y) 0 (- (+ root-y height) y)))))
                 ((= type xcb:ewmh:_NET_WM_MOVERESIZE_SIZE_TOPRIGHT)
                  (setq cursor exwm-floating--cursor-top-right
                        exwm-floating--moveresize-calculate
-                       `(lambda (x y)
-                          (vector ,buffer-or-id
-                                  ,(eval-when-compile
-                                     (logior xcb:ConfigWindow:Y
-                                             xcb:ConfigWindow:Width
-                                             xcb:ConfigWindow:Height))
-                                  0 (- y ,win-y) (- x ,(- root-x width))
-                                  (- ,(+ root-y height) y)))))
+                       (lambda (x y)
+                         (vector buffer-or-id
+                                 (eval-when-compile
+                                   (logior xcb:ConfigWindow:Y
+                                           xcb:ConfigWindow:Width
+                                           xcb:ConfigWindow:Height))
+                                 0 (- y win-y) (- x (- root-x width))
+                                 (- (+ root-y height) y)))))
                 ((= type xcb:ewmh:_NET_WM_MOVERESIZE_SIZE_RIGHT)
                  (setq cursor exwm-floating--cursor-right
                        exwm-floating--moveresize-calculate
-                       `(lambda (x y)
-                          (vector ,buffer-or-id
-                                  ,xcb:ConfigWindow:Width
-                                  0 0 (- x ,(- root-x width)) 0))))
+                       (lambda (x _y)
+                         (vector buffer-or-id
+                                 xcb:ConfigWindow:Width
+                                 0 0 (- x (- root-x width)) 0))))
                 ((= type xcb:ewmh:_NET_WM_MOVERESIZE_SIZE_BOTTOMRIGHT)
                  (setq cursor exwm-floating--cursor-bottom-right
                        exwm-floating--moveresize-calculate
-                       `(lambda (x y)
-                          (vector ,buffer-or-id
-                                  ,(eval-when-compile
-                                     (logior xcb:ConfigWindow:Width
-                                             xcb:ConfigWindow:Height))
-                                  0 0 (- x ,(- root-x width))
-                                  (- y ,(- root-y height))))))
+                       (lambda (x y)
+                         (vector buffer-or-id
+                                 (eval-when-compile
+                                   (logior xcb:ConfigWindow:Width
+                                           xcb:ConfigWindow:Height))
+                                 0 0 (- x (- root-x width))
+                                 (- y (- root-y height))))))
                 ((= type xcb:ewmh:_NET_WM_MOVERESIZE_SIZE_BOTTOM)
                  (setq cursor exwm-floating--cursor-bottom
                        exwm-floating--moveresize-calculate
-                       `(lambda (x y)
-                          (vector ,buffer-or-id
-                                  ,xcb:ConfigWindow:Height
-                                  0 0 0 (- y ,(- root-y height))))))
+                       (lambda (_x y)
+                         (vector buffer-or-id
+                                 xcb:ConfigWindow:Height
+                                 0 0 0 (- y (- root-y height))))))
                 ((= type xcb:ewmh:_NET_WM_MOVERESIZE_SIZE_BOTTOMLEFT)
                  (setq cursor exwm-floating--cursor-bottom-left
                        exwm-floating--moveresize-calculate
-                       `(lambda (x y)
-                          (vector ,buffer-or-id
-                                  ,(eval-when-compile
-                                     (logior xcb:ConfigWindow:X
-                                             xcb:ConfigWindow:Width
-                                             xcb:ConfigWindow:Height))
-                                  (- x ,win-x)
-                                  0
-                                  (- ,(+ root-x width) x)
-                                  (- y ,(- root-y height))))))
+                       (lambda (x y)
+                         (vector buffer-or-id
+                                 (eval-when-compile
+                                   (logior xcb:ConfigWindow:X
+                                           xcb:ConfigWindow:Width
+                                           xcb:ConfigWindow:Height))
+                                 (- x win-x)
+                                 0
+                                 (- (+ root-x width) x)
+                                 (- y (- root-y height))))))
                 ((= type xcb:ewmh:_NET_WM_MOVERESIZE_SIZE_LEFT)
                  (setq cursor exwm-floating--cursor-left
                        exwm-floating--moveresize-calculate
-                       `(lambda (x y)
-                          (vector ,buffer-or-id
-                                  ,(eval-when-compile
-                                     (logior xcb:ConfigWindow:X
-                                             xcb:ConfigWindow:Width))
-                                  (- x ,win-x) 0 (- ,(+ root-x width) x) 0)))))
+                       (lambda (x _y)
+                         (vector buffer-or-id
+                                 (eval-when-compile
+                                   (logior xcb:ConfigWindow:X
+                                           xcb:ConfigWindow:Width))
+                                 (- x win-x) 0 (- (+ root-x width) x) 0)))))
           ;; Select events and change cursor (should always succeed)
           (xcb:+request-unchecked+reply exwm--connection
               (make-instance 'xcb:GrabPointer
@@ -590,11 +648,38 @@ This is also used by X window containers.")
   (xcb:+request exwm--connection
       (make-instance 'xcb:UngrabPointer :time xcb:Time:CurrentTime))
   (when exwm-floating--moveresize-calculate
-    (let (result buffer-or-id)
+    (let (result buffer-or-id outer-id container-id)
       (setq result (funcall exwm-floating--moveresize-calculate 0 0)
             buffer-or-id (aref result 0))
       (when (bufferp buffer-or-id)
         (with-current-buffer buffer-or-id
+          (setq outer-id (frame-parameter exwm--floating-frame 'exwm-outer-id)
+                container-id (frame-parameter exwm--floating-frame
+                                              'exwm-container))
+          (with-slots (x y width height border-width)
+              (xcb:+request-unchecked+reply exwm--connection
+                  (make-instance 'xcb:GetGeometry
+                                 :drawable container-id))
+            ;; Notify Emacs frame about this the position change.
+            (xcb:+request exwm--connection
+                (make-instance 'xcb:SendEvent
+                               :propagate 0
+                               :destination outer-id
+                               :event-mask xcb:EventMask:StructureNotify
+                               :event
+                               (xcb:marshal
+                                (make-instance 'xcb:ConfigureNotify
+                                               :event outer-id
+                                               :window outer-id
+                                               :above-sibling xcb:Window:None
+                                               :x (+ x border-width)
+                                               :y (+ y border-width)
+                                               :width width
+                                               :height height
+                                               :border-width 0
+                                               :override-redirect 0)
+                                exwm--connection)))
+            (xcb:flush exwm--connection))
           (exwm-layout--show exwm--id
                              (frame-root-window exwm--floating-frame)))))
     (setq exwm-floating--moveresize-calculate nil)))
@@ -613,13 +698,18 @@ This is also used by X window containers.")
             y (aref result 3)
             width (max 1 (aref result 4))
             height (max 1 (aref result 5)))
-      (setq container-or-id
-            (if (bufferp buffer-or-id)
-                ;; Managed.
-                (with-current-buffer buffer-or-id
-                  (frame-parameter exwm--floating-frame 'exwm-container))
-              ;; Unmanaged.
-              buffer-or-id))
+      (if (not (bufferp buffer-or-id))
+          ;; Unmanaged.
+          (setq container-or-id buffer-or-id)
+        ;; Managed.
+        (setq container-or-id
+              (with-current-buffer buffer-or-id
+                (frame-parameter exwm--floating-frame 'exwm-container))
+              x (- x exwm-floating-border-width)
+              ;; Use `frame-outer-height' in the future.
+              y (- y exwm-floating-border-width
+                   exwm-workspace--window-y-offset)
+              height (+ height exwm-workspace--window-y-offset)))
       (xcb:+request exwm--connection
           (make-instance 'xcb:ConfigureWindow
                          :window container-or-id
@@ -668,32 +758,39 @@ Both DELTA-X and DELTA-Y default to 1.  This command should be bound locally."
                           nil nil))
     (xcb:flush exwm--connection)))
 
+(defun exwm-floating--init-border ()
+  "Initialize border colormap and pixel."
+  (exwm--log)
+  ;; Use the default colormap.
+  (unless exwm-floating--border-colormap
+    (with-slots (roots) (xcb:get-setup exwm--connection)
+      (with-slots (default-colormap) (car roots)
+        (setq exwm-floating--border-colormap default-colormap))))
+  ;; Free any previously allocated pixel.
+  (when exwm-floating--border-pixel
+    (xcb:+request exwm--connection
+        (make-instance 'xcb:FreeColors
+                       :cmap exwm-floating--border-colormap
+                       :plane-mask 0
+                       :pixels (vector exwm-floating--border-pixel)))
+    (setq exwm-floating--border-pixel nil))
+  ;; Allocate new pixel.
+  (let ((color (x-color-values (or exwm-floating-border-color "")))
+         reply)
+    (when color
+      (setq reply (xcb:+request-unchecked+reply exwm--connection
+                      (make-instance 'xcb:AllocColor
+                                     :cmap exwm-floating--border-colormap
+                                     :red (pop color)
+                                     :green (pop color)
+                                     :blue (pop color))))
+      (when reply
+        (setq exwm-floating--border-pixel (slot-value reply 'pixel))))))
+
 (defun exwm-floating--init ()
   "Initialize floating module."
   (exwm--log)
-  ;; Check border width.
-  (unless (and (integerp exwm-floating-border-width)
-               (> exwm-floating-border-width 0))
-    (setq exwm-floating-border-width 0))
-  ;; Initialize border pixel.
-  (when (> exwm-floating-border-width 0)
-    (setq exwm-floating--border-colormap
-          (slot-value (car (slot-value
-                            (xcb:get-setup exwm--connection) 'roots))
-                      'default-colormap))
-    (unless (stringp exwm-floating-border-color)
-      (setq exwm-floating-border-color ""))
-    (let* ((color (x-color-values exwm-floating-border-color))
-           reply)
-      (when color
-        (setq reply (xcb:+request-unchecked+reply exwm--connection
-                        (make-instance 'xcb:AllocColor
-                                       :cmap exwm-floating--border-colormap
-                                       :red (pop color)
-                                       :green (pop color)
-                                       :blue (pop color))))
-        (when reply
-          (setq exwm-floating--border-pixel (slot-value reply 'pixel))))))
+  (exwm-floating--init-border)
   ;; Initialize cursors for moving/resizing a window
   (xcb:cursor:init exwm--connection)
   (setq exwm-floating--cursor-move
