@@ -1,7 +1,7 @@
 ;;; exwm-manage.el --- Window Management Module for  -*- lexical-binding: t -*-
 ;;;                    EXWM
 
-;; Copyright (C) 2015-2018 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2019 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -48,32 +48,92 @@ You can still make the X windows floating afterwards."
   :type 'integer)
 
 (defcustom exwm-manage-configurations nil
-  "Per-application configurations."
+  "Per-application configurations.
+
+Configuration options allow to override various default behaviors of EXWM
+and only take effect when they are present.  Note for certain options
+specifying nil is not exactly the same as leaving them out.  Currently
+possible choices:
+* floating: Force floating (non-nil) or tiling (nil) on startup.
+* x/y/width/height: Override the initial geometry (floating X window only).
+* border-width: Override the border width (only visible when floating).
+* fullscreen: Force full screen (non-nil) on startup.
+* floating-mode-line: `mode-line-format' used when floating.
+* tiling-mode-line: `mode-line-format' used when tiling.
+* floating-header-line: `header-line-format' used when floating.
+* tiling-header-line: `header-line-format' used when tiling.
+* char-mode: Force char-mode (non-nil) on startup.
+* prefix-keys: `exwm-input-prefix-keys' local to this X window.
+* simulation-keys: `exwm-input-simulation-keys' local to this X window.
+* workspace: The initial workspace.
+* managed: Force to manage (non-nil) or not manage (nil) the X window.
+
+For each X window managed for the first time, matching criteria (sexps) are
+evaluated sequentially and the first configuration with a non-nil matching
+criterion would be applied.  Apart from generic forms, one would typically
+want to match against EXWM internal variables such as `exwm-title',
+`exwm-class-name' and `exwm-instance-name'."
   :type '(alist :key-type (sexp :tag "Matching criterion" nil)
                 :value-type
                 (plist :tag "Configurations"
-                       :key-type
-                       (choice
-                        (const :tag "Floating" floating)
-                        (const :tag "X" x)
-                        (const :tag "Y" y)
-                        (const :tag "Width" width)
-                        (const :tag "Height" height)
-                        (const :tag "Border width" border-width)
-                        (const :tag "Fullscreen" fullscreen)
-                        (const :tag "Floating mode-line" floating-mode-line)
-                        (const :tag "Tiling mode-line" tiling-mode-line)
-                        (const :tag "Floating header-line"
-                               floating-header-line)
-                        (const :tag "Tiling header-line" tiling-header-line)
-                        (const :tag "Char-mode" char-mode)
-                        (const :tag "Prefix keys" prefix-keys)
-                        (const :tag "Simulation keys" simulation-keys)
-                        (const :tag "Workspace" workspace)
-                        (const :tag "Managed" managed)
+                       :options
+                       (((const :tag "Floating" floating) boolean)
+                        ((const :tag "X" x) number)
+                        ((const :tag "Y" y) number)
+                        ((const :tag "Width" width) number)
+                        ((const :tag "Height" height) number)
+                        ((const :tag "Border width" border-width) integer)
+                        ((const :tag "Fullscreen" fullscreen) boolean)
+                        ((const :tag "Floating mode-line" floating-mode-line)
+                         sexp)
+                        ((const :tag "Tiling mode-line" tiling-mode-line) sexp)
+                        ((const :tag "Floating header-line"
+                                floating-header-line)
+                         sexp)
+                        ((const :tag "Tiling header-line" tiling-header-line)
+                         sexp)
+                        ((const :tag "Char-mode" char-mode) boolean)
+                        ((const :tag "Prefix keys" prefix-keys)
+                         (repeat key-sequence))
+                        ((const :tag "Simulation keys" simulation-keys)
+                         (alist :key-type (key-sequence :tag "From")
+                                :value-type (key-sequence :tag "To")))
+                        ((const :tag "Workspace" workspace) integer)
+                        ((const :tag "Managed" managed) boolean)
                         ;; For forward compatibility.
-                        (other))
-                       :value-type (sexp :tag "Value" nil))))
+                        ((other) sexp))))
+  ;; TODO: This is admittedly ugly.  We'd be better off with an event type.
+  :get (lambda (symbol)
+         (mapcar (lambda (pair)
+                   (let* ((match (car pair))
+                          (config (cdr pair))
+                          (prefix-keys (plist-get config 'prefix-keys)))
+                     (when prefix-keys
+                       (setq config (copy-tree config)
+                             config (plist-put config 'prefix-keys
+                                               (mapcar (lambda (i)
+                                                         (if (sequencep i)
+                                                             i
+                                                           (vector i)))
+                                                       prefix-keys))))
+                     (cons match config)))
+                 (default-value symbol)))
+  :set (lambda (symbol value)
+         (set symbol
+              (mapcar (lambda (pair)
+                        (let* ((match (car pair))
+                               (config (cdr pair))
+                               (prefix-keys (plist-get config 'prefix-keys)))
+                          (when prefix-keys
+                            (setq config (copy-tree config)
+                                  config (plist-put config 'prefix-keys
+                                                    (mapcar (lambda (i)
+                                                              (if (sequencep i)
+                                                                  (aref i 0)
+                                                                i))
+                                                            prefix-keys))))
+                          (cons match config)))
+                      value))))
 
 ;; FIXME: Make the following values as small as possible.
 (defconst exwm-manage--height-delta-min 5)
@@ -91,6 +151,7 @@ You can still make the X windows floating afterwards."
 (defvar exwm-manage--ping-lock nil
   "Non-nil indicates EXWM is pinging a window.")
 
+(defvar exwm-input--skip-buffer-list-update)
 (defvar exwm-input-prefix-keys)
 (defvar exwm-workspace--current)
 (defvar exwm-workspace--id-struts-alist)
@@ -183,7 +244,8 @@ You can still make the X windows floating afterwards."
   (when (derived-mode-p 'exwm-mode)
     (dolist (i exwm-manage-configurations)
       (save-current-buffer
-        (when (eval (car i) t)
+        (when (with-demoted-errors "Problematic configuration: %S"
+                (eval (car i) t))
           (cl-return-from exwm-manage--get-configurations (cdr i)))))))
 
 (defun exwm-manage--manage-window (id)
@@ -194,14 +256,15 @@ You can still make the X windows floating afterwards."
     (when (xcb:+request-checked+request-check exwm--connection
               (make-instance 'xcb:ChangeWindowAttributes
                              :window id :value-mask xcb:CW:EventMask
-                             :event-mask exwm--client-event-mask))
+                             :event-mask (exwm--get-client-event-mask)))
       (throw 'return 'dead))
     ;; Add this X window to save-set.
     (xcb:+request exwm--connection
         (make-instance 'xcb:ChangeSaveSet
                        :mode xcb:SetMode:Insert
                        :window id))
-    (with-current-buffer (generate-new-buffer "*EXWM*")
+    (with-current-buffer (let ((exwm-input--skip-buffer-list-update t))
+                           (generate-new-buffer "*EXWM*"))
       ;; Keep the oldest X window first.
       (setq exwm--id-buffer-alist
             (nconc exwm--id-buffer-alist `((,id . ,(current-buffer)))))
@@ -215,6 +278,8 @@ You can still make the X windows floating afterwards."
       (exwm--update-hints id)
       (exwm-manage--update-geometry id)
       (exwm-manage--update-mwm-hints id)
+      (exwm--update-title id)
+      (exwm--update-protocols id)
       (setq exwm--configurations (exwm-manage--get-configurations))
       ;; OverrideRedirect is not checked here.
       (when (and
@@ -253,7 +318,7 @@ You can still make the X windows floating afterwards."
                                      exwm-window-type)
                                ;; Listen for PropertyChange (struts) and
                                ;; UnmapNotify/DestroyNotify event of the dock.
-                               exwm--client-event-mask
+                               (exwm--get-client-event-mask)
                              xcb:EventMask:NoEvent)))
         ;; The window needs to be mapped
         (xcb:+request exwm--connection
@@ -284,7 +349,8 @@ You can still make the X windows floating afterwards."
                              :stack-mode xcb:StackMode:Below)))
         (xcb:flush exwm--connection)
         (setq exwm--id-buffer-alist (assq-delete-all id exwm--id-buffer-alist))
-        (let ((kill-buffer-query-functions nil))
+        (let ((kill-buffer-query-functions nil)
+              (exwm-input--skip-buffer-list-update t))
           (kill-buffer (current-buffer)))
         (throw 'return 'ignored))
       (let ((index (plist-get exwm--configurations 'workspace)))
@@ -308,8 +374,6 @@ You can still make the X windows floating afterwards."
                            :button button :modifiers xcb:ModMask:Any)))
       (exwm-manage--set-client-list)
       (xcb:flush exwm--connection)
-      (exwm--update-title id)
-      (exwm--update-protocols id)
       (if (plist-member exwm--configurations 'floating)
           ;; User has specified whether it should be floating.
           (if (plist-get exwm--configurations 'floating)
@@ -452,8 +516,10 @@ manager is shutting down."
   (catch 'return
     (when (or (not exwm--id)
               (xcb:+request-checked+request-check exwm--connection
-                  (make-instance 'xcb:MapWindow
-                                 :window exwm--id)))
+                  (make-instance 'xcb:ChangeWindowAttributes
+                                 :window exwm--id
+                                 :value-mask xcb:CW:EventMask
+                                 :event-mask (exwm--get-client-event-mask))))
       ;; The X window is no longer alive so just close the buffer.
       (when exwm--floating-frame
         (let ((window (frame-parameter exwm--floating-frame 'exwm-outer-id))
@@ -537,8 +603,8 @@ Would you like to kill it? "
       (signal-process pid 'SIGKILL)
       ;; Ensure it's dead
       (run-with-timer exwm-manage-ping-timeout nil
-                      `(lambda ()
-                         (xcb:+request exwm--connection ,request))))
+                      (lambda ()
+                        (xcb:+request exwm--connection request))))
     (xcb:flush exwm--connection)))
 
 (defun exwm-manage--add-frame (frame)
@@ -712,14 +778,7 @@ border-width: %d; sibling: #x%x; stack-mode: %d"
   "Initialize manage module."
   ;; Intern _MOTIF_WM_HINTS
   (exwm--log)
-  (let ((atom-name "_MOTIF_WM_HINTS"))
-    (setq exwm-manage--_MOTIF_WM_HINTS
-          (slot-value (xcb:+request-unchecked+reply exwm--connection
-                          (make-instance 'xcb:InternAtom
-                                         :only-if-exists 0
-                                         :name-len (length atom-name)
-                                         :name atom-name))
-                      'atom)))
+  (setq exwm-manage--_MOTIF_WM_HINTS (exwm--intern-atom "_MOTIF_WM_HINTS"))
   (add-hook 'after-make-frame-functions #'exwm-manage--add-frame)
   (add-hook 'delete-frame-functions #'exwm-manage--remove-frame)
   (xcb:+event exwm--connection 'xcb:ConfigureRequest
