@@ -1,11 +1,11 @@
 ;;; exwm.el --- Emacs X Window Manager  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2015-2018 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2019 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 ;; Maintainer: Chris Feng <chris.w.feng@gmail.com>
-;; Version: 0.21
-;; Package-Requires: ((xelb "0.16"))
+;; Version: 0.23
+;; Package-Requires: ((xelb "0.18"))
 ;; Keywords: unix
 ;; URL: https://github.com/ch11ng/exwm
 
@@ -99,6 +99,12 @@
   "Subrs (primitives) that would normally block EXWM."
   :type '(repeat function))
 
+(defcustom exwm-replace 'ask
+  "Whether to replace existing window manager."
+  :type '(radio (const :tag "Ask" ask)
+                (const :tag "Replace by default" t)
+                (const :tag "Do not replace" nil)))
+
 (defconst exwm--server-name "server-exwm"
   "Name of the subordinate Emacs server.")
 
@@ -414,7 +420,8 @@
     (setq type (slot-value obj 'type)
           id (slot-value obj 'window)
           data (slot-value (slot-value obj 'data) 'data32))
-    (exwm--log "atom=%s(%s)" (x-get-atom-name type exwm-workspace--current) type)
+    (exwm--log "atom=%s(%s)" (x-get-atom-name type exwm-workspace--current)
+               type)
     (cond
      ;; _NET_NUMBER_OF_DESKTOPS.
      ((= type xcb:Atom:_NET_NUMBER_OF_DESKTOPS)
@@ -426,7 +433,9 @@
           (make-frame))
          ((and (> current requested)
                (> current 1))
-          (delete-frame (car (last exwm-workspace--list)))))))
+          (let ((frame (car (last exwm-workspace--list))))
+            (exwm-workspace--get-remove-frame-next-workspace frame)
+            (delete-frame frame))))))
      ;; _NET_CURRENT_DESKTOP.
      ((= type xcb:Atom:_NET_CURRENT_DESKTOP)
       (exwm-workspace-switch (elt data 0)))
@@ -451,7 +460,7 @@
      ((= type xcb:Atom:_NET_CLOSE_WINDOW)
       (let ((buffer (exwm--id->buffer id)))
         (when (buffer-live-p buffer)
-          (kill-buffer buffer))))
+          (exwm--defer 0 #'kill-buffer buffer))))
      ;; _NET_WM_MOVERESIZE
      ((= type xcb:Atom:_NET_WM_MOVERESIZE)
       (let ((direction (elt data 2))
@@ -549,7 +558,9 @@
         (when (and (buffer-live-p buffer)
                    (= (elt data 0) xcb:icccm:WM_STATE:IconicState))
           (with-current-buffer buffer
-            (bury-buffer)))))
+            (if exwm--floating-frame
+                (call-interactively #'exwm-floating-hide)
+              (bury-buffer))))))
      (t
       (exwm--log "Unhandled: %s(%d)"
                  (x-get-atom-name type exwm-workspace--current) type)))))
@@ -721,7 +732,7 @@ manager.  If t, replace it, if nil, abort and ask the user if `ask'."
       (when (eq replace 'ask)
         (setq replace (yes-or-no-p "Replace existing window manager? ")))
       (when (not replace)
-        (error "Other window manager detected")))
+        (user-error "Other window manager detected")))
     (let ((new-owner (xcb:generate-id exwm--connection)))
       (xcb:+request exwm--connection
           (make-instance 'xcb:CreateWindow
@@ -812,7 +823,7 @@ manager.  If t, replace it, if nil, abort and ask the user if `ask'."
         (xcb:icccm:init exwm--connection t)
         (xcb:ewmh:init exwm--connection t)
         ;; Try to register window manager selection.
-        (exwm--wmsn-acquire 'ask)
+        (exwm--wmsn-acquire exwm-replace)
         (when (xcb:+request-checked+request-check exwm--connection
                   (make-instance 'xcb:ChangeWindowAttributes
                                  :window exwm--root
@@ -833,9 +844,10 @@ manager.  If t, replace it, if nil, abort and ask the user if `ask'."
         (exwm--unlock)
         (exwm-workspace--post-init)
         (exwm-input--post-init)
+        (run-hooks 'exwm-init-hook)
         ;; Manage existing windows
-        (exwm-manage--scan)
-        (run-hooks 'exwm-init-hook))
+        (exwm-manage--scan))
+    (user-error)
     ((quit error)
      (exwm-exit)
      ;; Rethrow error
@@ -860,6 +872,7 @@ manager.  If t, replace it, if nil, abort and ask the user if `ask'."
     (xcb:disconnect exwm--connection))
   (setq exwm--connection nil))
 
+;;;###autoload
 (defun exwm-enable (&optional undo)
   "Enable/Disable EXWM."
   (exwm--log "%s" undo)
@@ -879,10 +892,11 @@ manager.  If t, replace it, if nil, abort and ask the user if `ask'."
      ;; Ignore unrecognized command line arguments.  This can be helpful
      ;; when EXWM is launched by some session manager.
      (push #'vector command-line-functions)
-     (add-hook (if (daemonp)
-                   'after-make-frame-functions ;emacsclient
-                 'window-setup-hook)           ;emacs
-               #'exwm-init t)
+     ;; In case EXWM is to be started from a graphical Emacs instance.
+     (add-hook 'window-setup-hook #'exwm-init t)
+     ;; In case EXWM is to be started with emacsclient.
+     (add-hook 'after-make-frame-functions #'exwm-init t)
+     ;; Manage the subordinate Emacs server.
      (add-hook 'kill-emacs-hook #'exwm--server-stop)
      (dolist (i exwm-blocking-subrs)
        (advice-add i :around #'exwm--server-eval-at)))))
